from abc import ABC, abstractmethod


class CustomFunctionInterface(ABC):
    @property
    @abstractmethod
    def name(self) -> str:
        "Custom function name"
        raise NotImplementedError



"""Service access layer exceptions."""


class ServiceAccessError(Exception):
    pass


class CommunicationError(ServiceAccessError):
    pass


class SSLError(ServiceAccessError):
    pass


class InvalidContentTypeError(ServiceAccessError):
    pass


class UnauthorizedError(ServiceAccessError):
    pass


class HTTPError(ServiceAccessError):
    def __init__(self, message: str, status_code: int):
        self.status_code = status_code
        super().__init__(message)


class ErrorBodyIsNotJSONError(ServiceAccessError):
    pass


class ErrorJSONInvalidFormatError(ServiceAccessError):
    pass


class NoResponseFetchVersionError(ServiceAccessError):
    pass



"""This module contains operators for parametrizing batch runs."""
from abc import abstractmethod
from dataclasses import dataclass
from typing import Any


class Operator:
    """Base class for an Operator."""

    @abstractmethod
    def __str__(self) -> str:
        "Returns a string representation of the operator"


@dataclass
class Range(Operator):
    """Range operator class for parametrizing batch runs. Range(a, b, c) represents c
    linearly spaced values in the (real) interval [a, b]; b >= a. The functionality is
    analogous to numpy.linspace.

    Args:
        start_value: The start value for the sweep parameter.
        end_value: The end value for the sweep parameter.
        no_of_steps: The number of steps to intermediate steps
        to take between start_value and end_value.

    Example::

        from modelon.impact.client import Range

        fmu = model.compile().wait()
        experiment_definition = fmu.new_experiment_definition(
            custom_function).with_modifiers({'inertia1.J': 2,
            'inertia2.J': Range(0.1, 0.5, 3)})

    """

    start_value: float
    end_value: float
    no_of_steps: int

    def __str__(self) -> str:
        return f"range({self.start_value},{self.end_value},{self.no_of_steps})"


class Choices(Operator):
    """Choices operator class for parametrizing batch runs. Choices defines a list of
    specified values a Modifier expression can take.

    Args:
        values: Variable number of numerical arguments to sweep.

    Example::

        from modelon.impact.client import Choices

        fmu = model.compile().wait()
        experiment_definition = fmu.new_experiment_definition(
            custom_function).with_modifiers({'inertia1.J': 2,
            'inertia2.J': Choices(0.1, 0.5)})

    """

    def __init__(self, *values: Any):
        self.values = values

    def __str__(self) -> str:
        return f"choices({', '.join(map(str, self.values))})"


@dataclass
class Uniform(Operator):
    """Uniform distribution class for parametrizing batch runs.For mathematical
    background, see e.g.,
    https://en.wikipedia.org/wiki/Continuous_uniform_distribution.

    Args:
        start (float): Starting value of the interval.
        end (float): End value of the interval. Requires start <= end.

    Example::

        from modelon.impact.client import Uniform

        fmu = model.compile().wait()
        experiment_definition = fmu.new_experiment_definition(
            custom_function).with_modifiers({'inertia1.J': 2,
            'inertia2.J': Uniform(0.1, 0.5)})

    """

    start: float
    end: float

    def __str__(self) -> str:
        return f"uniform({self.start},{self.end})"


@dataclass
class Beta(Operator):
    """Beta distribution class for parametrizing batch runs. For mathematical
    background, see e.g., https://en.wikipedia.org/wiki/Beta_distribution.

    Args:
        alpha (float): 'alpha' resp. 'a' parameter of beta distribution, requires
            alpha > 0
        beta (float): 'beta' resp. 'b' parameter of beta distribution, requires
            beta > 0

    Example::

        from modelon.impact.client import Beta

        fmu = model.compile().wait()
        experiment_definition = fmu.new_experiment_definition(
            custom_function).with_modifiers({'inertia1.J': 2,
            'inertia2.J': Beta(0.1, 0.5, 3)})

    """

    alpha: float
    beta: float

    def __str__(self) -> str:
        return f"beta({self.alpha},{self.alpha})"


@dataclass
class Normal(Operator):
    """Normal distribution class for parametrizing batch runs.For mathematical
    background, see e.g., https://en.wikipedia.org/wiki/Normal_distribution
    https://en.wikipedia.org/wiki/Truncated_normal_distribution Supports both
    the standard and truncated Normal distribution. The standard Normal
    distribution is the default, add additional start & end parameters for
    truncation.

    Args:
        mean (float): Mean resp. location of the distribution.
        variance (float): Variance of the normal distribution. Requires var > 0.
        start (float): Lower bound. Default: -inf
        end (float): Upper bound, requires end > start. Default: inf

    Example::

        from modelon.impact.client import Normal

        fmu = model.compile().wait()
        experiment_definition = fmu.new_experiment_definition(
            custom_function).with_modifiers({'inertia1.J': 2,
            'inertia2.J': Normal(0.1, 0.5)})

    """

    mean: float
    variance: float
    start: float = float("-inf")
    end: float = float("inf")

    def __str__(self) -> str:
        return f"normal({self.mean},{self.variance},{self.start},{self.end})"



"""Service class."""
import inspect
import logging
from typing import Any, Callable, Dict, Optional, Union

from modelon.impact.client.jupyterhub.sal import JupyterContext
from modelon.impact.client.sal import exceptions
from modelon.impact.client.sal.context import Context
from modelon.impact.client.sal.custom_function import CustomFunctionService
from modelon.impact.client.sal.experiment import ExperimentService
from modelon.impact.client.sal.exports import ExportService
from modelon.impact.client.sal.external_result import ExternalResultService
from modelon.impact.client.sal.http import HTTPClient
from modelon.impact.client.sal.imports import ImportService
from modelon.impact.client.sal.model_executable import ModelExecutableService
from modelon.impact.client.sal.project import ProjectService
from modelon.impact.client.sal.uri import URI
from modelon.impact.client.sal.users import UsersService
from modelon.impact.client.sal.workspace import WorkspaceService

logger = logging.getLogger(__name__)


def _decorate_all_methods(cls: Any, decorator: Callable) -> Any:
    for method_name, method in inspect.getmembers(cls, lambda x: inspect.ismethod(x)):
        setattr(cls, method_name, decorator(method))

    return cls


class Service:
    _JUPYTERHUB_VERSION_HEADER = "x-jupyterhub-version"

    def __init__(
        self, uri: URI, context: Optional[Union[Context, JupyterContext]] = None
    ):
        self._base_uri = uri
        self._http_client = HTTPClient(context)
        self.workspace = WorkspaceService(self._base_uri, self._http_client)
        self.project = ProjectService(self._base_uri, self._http_client)
        self.model_executable = ModelExecutableService(
            self._base_uri, self._http_client
        )
        self.experiment = ExperimentService(self._base_uri, self._http_client)
        self.custom_function = CustomFunctionService(self._base_uri, self._http_client)
        self.external_result = ExternalResultService(self._base_uri, self._http_client)
        self.users = UsersService(self._base_uri, self._http_client)
        self.exports = ExportService(self._base_uri, self._http_client)
        self.imports = ImportService(self._base_uri, self._http_client)

    def add_login_retry_with(self, api_key: Optional[str] = None) -> None:
        def retry_with_login_decorator(func: Callable) -> Callable:
            def wrapped(*args: Any, **kwargs: Any) -> Callable:
                try:
                    return func(*args, **kwargs)
                except exceptions.HTTPError as e:
                    if e.status_code != 401:
                        raise

                    self.api_login(api_key=api_key)
                    return func(*args, **kwargs)

            return wrapped

        self.workspace = _decorate_all_methods(
            self.workspace, retry_with_login_decorator
        )
        self.project = _decorate_all_methods(self.project, retry_with_login_decorator)
        self.model_executable = _decorate_all_methods(
            self.model_executable, retry_with_login_decorator
        )
        self.experiment = _decorate_all_methods(
            self.experiment, retry_with_login_decorator
        )
        self.custom_function = _decorate_all_methods(
            self.custom_function, retry_with_login_decorator
        )
        self.users = _decorate_all_methods(self.users, retry_with_login_decorator)
        self.exports = _decorate_all_methods(self.exports, retry_with_login_decorator)
        self.imports = _decorate_all_methods(self.imports, retry_with_login_decorator)
        self.external_result = _decorate_all_methods(
            self.external_result, retry_with_login_decorator
        )

    def is_jupyterhub_url(self) -> bool:
        url = (self._base_uri / "hub/api/").resolve()

        try:
            response = self._http_client.get_json_response(url)
        except (
            exceptions.CommunicationError,
            exceptions.SSLError,
            exceptions.HTTPError,
            exceptions.ErrorBodyIsNotJSONError,
        ):
            return False
        except Exception:
            logger.warning(
                "Unknown exception trying to determine if URL is to JupyterHub or "
                "Modelon Impact, will assume the URL goes directly to the "
                "Modelon Impact API"
            )
            return False
        return self._JUPYTERHUB_VERSION_HEADER in response.headers

    def api_get_metadata(self) -> Dict[str, Any]:
        url = (self._base_uri / "api/").resolve()
        response = self._http_client.get_json_response(url)

        return response.data

    def api_login(self, api_key: Optional[str] = None) -> Dict[str, Any]:
        login_data = {"secretKey": api_key} if api_key else {}
        url = (self._base_uri / "api/login").resolve()
        return self._http_client.post_json(url, login_data)

    def get_executions(self) -> Dict[str, Any]:
        url = (self._base_uri / "api/executions").resolve()
        resp = self._http_client.get_json_response(url)
        return resp.data



from abc import ABC, abstractmethod
from typing import Any


class ExperimentInterface(ABC):
    @property
    @abstractmethod
    def id(self) -> str:
        """Experiment id."""
        raise NotImplementedError

    @abstractmethod
    def get_cases(self) -> Any:
        """Returns a list of case objects for an experiment."""
        raise NotImplementedError



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict

from modelon.impact.client import exceptions
from modelon.impact.client.operations.base import (
    AsyncOperation,
    AsyncOperationStatus,
    Entity,
)
from modelon.impact.client.sal.service import Service

if TYPE_CHECKING:
    from modelon.impact.client.operations.base import EntityFromOperation


class ProjectImportOperation(AsyncOperation[Entity]):
    """An import operation class for the Project class."""

    def __init__(
        self, location: str, service: Service, create_entity: EntityFromOperation
    ):
        super().__init__(create_entity)
        self._location = location
        self._sal = service
        self._create_entity = create_entity

    def __repr__(self) -> str:
        return f"Project import operations for id '{self.id}'"

    def __eq__(self, obj: object) -> bool:
        return (
            isinstance(obj, ProjectImportOperation) and obj._location == self._location
        )

    @property
    def id(self) -> str:
        """Project import id."""
        return self._location.split("/")[-1]

    @property
    def name(self) -> str:
        """Return the name of operation."""
        return "Project import"

    def _info(self) -> Dict[str, Any]:
        return self._sal.imports.get_import_status(self._location)["data"]

    def data(self) -> Entity:
        """Returns a new Project class instance.

        Returns:
            A Project class instance.

        """
        info = self._info()
        if info["status"] == AsyncOperationStatus.ERROR.value:
            raise exceptions.IllegalProjectImport(
                f"Project import failed! Cause: {info['error'].get('message')}"
            )
        project_id = info["data"]["projectId"]
        resp = self._sal.project.project_get(project_id, False, False)
        return self._create_entity(
            self,
            project_id=resp["id"],
            project_definition=resp["definition"],
            project_type=resp["projectType"],
            storage_location=resp["storageLocation"],
        )

    @property
    def status(self) -> AsyncOperationStatus:
        """Returns the upload status as an enumeration.

        Returns:
            The AsyncOperationStatus enum. The status can have the enum values
            AsyncOperationStatus.READY, AsyncOperationStatus.RUNNING or
            AsyncOperationStatus.ERROR

        Example::

            status = client.import_project_from_zip(path_to_project).status
            status = workspace.import_project_from_zip(path_to_project).status

        """
        return AsyncOperationStatus(self._info()["status"])



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Dict, Optional, Union

from modelon.impact.client.entities.interfaces.case import CaseInterface
from modelon.impact.client.entities.interfaces.experiment import ExperimentInterface
from modelon.impact.client.experiment_definition.interfaces.extension import (
    BaseExperimentExtension,
)
from modelon.impact.client.experiment_definition.operators import Operator
from modelon.impact.client.experiment_definition.util import (
    case_to_identifier_dict,
    get_options,
)

if TYPE_CHECKING:
    from modelon.impact.client.entities.case import Case
    from modelon.impact.client.entities.experiment import Experiment

    CaseOrExperiment = Union[Case, Experiment]

logger = logging.getLogger(__name__)


def _validate_initialize_from(entity: Optional[CaseOrExperiment]) -> None:
    if not entity:
        return
    if entity and not isinstance(entity, (ExperimentInterface, CaseInterface)):
        raise TypeError(
            "It is only supported to specify initialize from either Experiment "
            "or Case for experiment extensions"
        )
    if isinstance(entity, ExperimentInterface) and len(entity.get_cases()) > 1:
        raise ValueError(
            "Cannot initialize from an experiment containing multiple"
            " cases! Please specify a case object instead."
        )


class SimpleExperimentExtension(BaseExperimentExtension):
    """A simple experiment extension class for defining experiment extensions.

    Args:
        parameter_modifiers: The custom function parameters passes as a dictionary.
            By default, the parameter_modifier is set to None, which means the options
            set in the experiment definition will be used.
        solver_options: A solver options class instance of SolverOptions or
            a dictionary object containing the solver options. By
            default, the options is set to None, which means an empty dictionary
            is passed in the experiment extension.
        simulation_options: A simulation options class instance of SimulationOptions or
            a dictionary object containing the simulation options. By
            default, the options is set to None, which means an empty dictionary
            is passed in the experiment extension.
        simulation_log_level: Simulation log level for this experiment.
            Default: 'WARNING'.
        initialize_from: Optional entity to initialize from. An instance of
            Case or Experiment. Default: None

    Example::

        fmu = model.compile().wait()
        simulation_options = custom_function.get_simulation_options()
        .with_values(ncp=500)
        solver_options = {'atol':1e-8}
        simulate_def = fmu.new_experiment_definition(
            custom_function,
            solver_options=solver_options,
            simulation_options=simulation_options
        ).with_modifiers({'inertia1.J': 2})
        simulate_ext = SimpleExperimentExtension(
        {'start_time': 0.0, 'final_time': 4.0},
        solver_options,
        simulation_options.with_values(ncp=600)
        ).with_modifiers({'PI.k': 40})
        simulate_def = simulate_def.with_extensions(simulate_ext)
        simulate_def.to_dict()

    """

    def __init__(
        self,
        parameter_modifiers: Optional[Dict[str, Any]] = None,
        solver_options: Optional[Dict[str, Any]] = None,
        simulation_options: Optional[Dict[str, Any]] = None,
        simulation_log_level: Optional[str] = None,
        initialize_from: Optional[CaseOrExperiment] = None,
    ):
        _validate_initialize_from(initialize_from)
        self._parameter_modifiers = (
            {} if parameter_modifiers is None else parameter_modifiers
        )
        self._solver_options = get_options(dict, solver_options)
        self._simulation_options = get_options(dict, simulation_options)
        self._simulation_log_level = simulation_log_level
        self._variable_modifiers: Dict[str, Any] = {}
        self._initialize_from = initialize_from
        self._case_label: Optional[str] = None

    def with_modifiers(
        self, modifiers: Optional[Dict[str, Any]] = None, **modifiers_kwargs: Any
    ) -> SimpleExperimentExtension:
        """Sets the modifiers variables for an experiment extension.

        Args:
            modifiers: A dictionary of variable modifiers.

        Example::

            simulation_options = custom_function.get_simulation_options()
            .with_values(ncp=500)
            solver_options = {'atol':1e-8}
            simulate_ext = SimpleExperimentExtension().with_modifiers({'PI.k': 40})
            simulate_ext = SimpleExperimentExtension(
            {'start_time': 0.0, 'final_time': 4.0},
            solver_options,
            simulation_options
            ).with_modifiers({'PI.k': 40})

        """
        if modifiers_kwargs:
            logger.warning(
                "The argument 'modifier_kwargs' is deprecated and will be removed in a"
                " future release of the Impact Client. Please specify a "
                "dictionary of variable modifiers instead!"
            )
        modifiers = {} if modifiers is None else modifiers
        modifiers_aggregate = {**modifiers, **modifiers_kwargs}
        new = SimpleExperimentExtension(
            self._parameter_modifiers,
            self._solver_options,
            self._simulation_options,
            self._simulation_log_level,
            self._initialize_from,
        )
        new._case_label = self._case_label
        for variable, value in modifiers_aggregate.items():
            if isinstance(value, Operator):
                raise ValueError(
                    "Range operator is not supported when using extentions"
                    " in the experiment"
                )
            new._variable_modifiers[variable] = value
        return new

    def with_case_label(self, case_label: str) -> SimpleExperimentExtension:
        """Sets the case label for an experiment extension.

        Args:
            case_label: A case label string.

        Example::

            simulation_options = custom_function.get_simulation_options()
            .with_values(ncp=500)
            solver_options = {'atol':1e-8}
            simulate_ext = SimpleExperimentExtension().with_case_label(
            'Cruise condition')
            simulate_ext = SimpleExperimentExtension(
            {'start_time': 0.0, 'final_time': 4.0},
            solver_options,
            simulation_options
            ).with_case_label('Cruise condition')

        """
        new = SimpleExperimentExtension(
            self._parameter_modifiers,
            self._solver_options,
            self._simulation_options,
            self._simulation_log_level,
            self._initialize_from,
        )
        new._variable_modifiers = self._variable_modifiers
        new._case_label = case_label

        return new

    @property
    def initialize_from(self) -> Optional[CaseOrExperiment]:
        return self._initialize_from

    def with_initialize_from(
        self, entity: Optional[CaseOrExperiment] = None
    ) -> SimpleExperimentExtension:
        """Sets the experiment or case to initialize from for an experiment extension.

        Args:
            entity: An instance of Case or Experiment classes.

        Example::

            experiment = workspace.get_experiment(experiment_id)
            simulate_ext = SimpleExperimentExtension()
            simulate_ext.with_initialize_from(experiment)

            experiment = workspace.get_experiment(experiment_id)
            case = experiment.get_case('case_1')
            simulate_ext = SimpleExperimentExtension()
            simulate_ext.with_initialize_from(case)

        """
        _validate_initialize_from(entity)
        new = SimpleExperimentExtension(
            self._parameter_modifiers,
            self._solver_options,
            self._simulation_options,
            self._simulation_log_level,
            entity,
        )
        new._variable_modifiers = self._variable_modifiers
        new._case_label = self._case_label
        return new

    def to_dict(self) -> Dict[str, Any]:
        """Returns the experiment extensions as a dictionary.

        Returns:
            A dictionary containing the experiment extensions.

        Example::

            fmu = model.compile().wait()
            simulation_options = custom_function.get_simulation_options()
                .with_values(ncp=500)
            solver_options = {'atol':1e-8}
            simulate_ext = SimpleExperimentExtension(
            {'start_time': 0.0, 'final_time': 4.0},
            solver_options,
            simulation_options,
            ).with_modifiers({'PI.k': 40})
            simulate_ext.to_dict()

        """
        ext_dict: Dict[str, Any] = {}
        if self._variable_modifiers:
            ext_dict.setdefault("modifiers", {})
            ext_dict["modifiers"]["variables"] = self._variable_modifiers

        if self._parameter_modifiers:
            ext_dict.setdefault("analysis", {})
            ext_dict["analysis"]["parameters"] = self._parameter_modifiers

        if self._solver_options:
            ext_dict.setdefault("analysis", {})
            ext_dict["analysis"]["solverOptions"] = self._solver_options

        if self._simulation_options:
            ext_dict.setdefault("analysis", {})
            ext_dict["analysis"]["simulationOptions"] = self._simulation_options

        if self._simulation_log_level:
            ext_dict.setdefault("analysis", {})
            ext_dict["analysis"]["simulationLogLevel"] = self._simulation_log_level

        if isinstance(self.initialize_from, ExperimentInterface):
            ext_dict.setdefault("modifiers", {})
            ext_dict["modifiers"]["initializeFrom"] = self.initialize_from.id
        elif isinstance(self.initialize_from, CaseInterface):
            ext_dict["modifiers"]["initializeFromCase"] = case_to_identifier_dict(
                self.initialize_from
            )

        if self._case_label:
            ext_dict["caseData"] = [{"label": self._case_label}]

        return ext_dict



from __future__ import annotations

from typing import TYPE_CHECKING

from modelon.impact.client.operations.base import Entity, ExecutionOperation, Status

if TYPE_CHECKING:
    from modelon.impact.client.operations.base import EntityFromOperation
    from modelon.impact.client.sal.service import Service


class ExperimentOperation(ExecutionOperation[Entity]):
    """An operation class for the Experiment class."""

    def __init__(
        self,
        workspace_id: str,
        exp_id: str,
        service: Service,
        create_entity: EntityFromOperation,
    ):
        super().__init__(create_entity)
        self._workspace_id = workspace_id
        self._exp_id = exp_id
        self._sal = service
        self._create_entity = create_entity

    def __repr__(self) -> str:
        return f"Experiment operation for id '{self._exp_id}'"

    def __eq__(self, obj: object) -> bool:
        return isinstance(obj, ExperimentOperation) and obj._exp_id == self._exp_id

    @property
    def id(self) -> str:
        """Experiment id."""
        return self._exp_id

    @property
    def name(self) -> str:
        """Return the name of operation."""
        return "Execution"

    def data(self) -> Entity:
        """Returns a new Experiment class instance.

        Returns:
            An experiment class instance.

        """
        return self._create_entity(
            self, workspace_id=self._workspace_id, exp_id=self._exp_id
        )

    @property
    def status(self) -> Status:
        """Returns the execution status as an enumeration.

        Returns:
            The execution status enum. The status can have the enum values
            Status.PENDING, Status.RUNNING, Status.STOPPING, Status.CANCELLED
            or Status.DONE

        Example::

            workspace.execute(definition).status

        """
        return Status(
            self._sal.experiment.execute_status(self._workspace_id, self._exp_id)[
                "status"
            ]
        )

    def cancel(self) -> None:
        """Terminates the execution process.

        Example::

            workspace.execute(definition).cancel()

        """
        self._sal.experiment.execute_cancel(self._workspace_id, self._exp_id)



"""Custom function service module."""
from typing import Any, Dict

from modelon.impact.client.sal.http import HTTPClient
from modelon.impact.client.sal.uri import URI


class CustomFunctionService:
    def __init__(self, uri: URI, http_client: HTTPClient):
        self._base_uri = uri
        self._http_client = http_client

    def custom_function_get(
        self, workspace_id: str, custom_function: str
    ) -> Dict[str, Any]:
        url = (
            self._base_uri
            / f"api/workspaces/{workspace_id}/custom-functions/{custom_function}"
        ).resolve()
        return self._http_client.get_json(url)

    def custom_functions_get(self, workspace_id: str) -> Dict[str, Any]:
        url = (
            self._base_uri / f"api/workspaces/{workspace_id}/custom-functions"
        ).resolve()
        return self._http_client.get_json(url)

    def custom_function_default_options_get(
        self, workspace_id: str, custom_function: str
    ) -> Dict[str, Any]:
        url = (
            self._base_uri
            / f"api/workspaces/{workspace_id}/custom-functions/{custom_function}"
            "/default-options"
        ).resolve()
        return self._http_client.get_json(url)

    def custom_function_options_get(
        self, workspace_id: str, custom_function: str
    ) -> Dict[str, Any]:
        url = (
            self._base_uri
            / f"api/workspaces/{workspace_id}/custom-functions/{custom_function}"
            "/options"
        ).resolve()
        return self._http_client.get_json(url)



from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from modelon.impact.client.entities.interfaces.model import ModelInterface
from modelon.impact.client.entities.model_executable import ModelExecutable
from modelon.impact.client.entities.project import Project
from modelon.impact.client.experiment_definition.model_based import (
    SimpleModelicaExperimentDefinition,
)
from modelon.impact.client.operations.fmu_import import FMUImportOperation
from modelon.impact.client.operations.model_executable import (
    CachedModelExecutableOperation,
    ModelExecutableOperation,
)
from modelon.impact.client.options import (
    CompilerOptions,
    ProjectExecutionOptions,
    RuntimeOptions,
    SimulationOptions,
    SolverOptions,
)

if TYPE_CHECKING:
    from modelon.impact.client.entities.case import Case
    from modelon.impact.client.entities.custom_function import CustomFunction
    from modelon.impact.client.entities.experiment import Experiment
    from modelon.impact.client.entities.external_result import ExternalResult
    from modelon.impact.client.operations.base import BaseOperation
    from modelon.impact.client.sal.service import Service

    CaseOrExperimentOrExternalResult = Union[Case, Experiment, ExternalResult]

logger = logging.getLogger(__name__)

CompilationOperations = Union[ModelExecutableOperation, CachedModelExecutableOperation]

RuntimeOptionsOrDict = Union[RuntimeOptions, Dict[str, Any]]
CompilerOptionsOrDict = Union[CompilerOptions, Dict[str, Any]]
SimulationOptionsOrDict = Union[SimulationOptions, Dict[str, Any]]
SolverOptionsOrDict = Union[SolverOptions, Dict[str, Any]]


def _assert_valid_compilation_options(
    compiler_options: Optional[CompilerOptionsOrDict] = None,
    runtime_options: Optional[RuntimeOptionsOrDict] = None,
) -> None:
    if compiler_options is not None and not isinstance(
        compiler_options, (CompilerOptions, dict)
    ):
        raise TypeError(
            "Compiler options object must either be a dictionary or an "
            "instance of CompilerOptions class!"
        )
    if runtime_options is not None and not isinstance(
        runtime_options, (RuntimeOptions, dict)
    ):
        raise TypeError(
            "Runtime options object must either be a dictionary or an "
            "instance of RuntimeOptions class!"
        )


class Model(ModelInterface):
    """Class containing Model functionalities."""

    def __init__(
        self, class_name: str, workspace_id: str, project_id: str, service: Service
    ):
        self._class_name = class_name
        self._workspace_id = workspace_id
        self._project_id = project_id
        self._sal = service

    def __repr__(self) -> str:
        return f"Class name '{self._class_name}'"

    def __eq__(self, obj: object) -> bool:
        return isinstance(obj, Model) and obj._class_name == self._class_name

    @property
    def name(self) -> str:
        """Class name."""
        return self._class_name

    def compile(
        self,
        compiler_options: CompilerOptionsOrDict,
        runtime_options: Optional[RuntimeOptionsOrDict] = None,
        compiler_log_level: str = "warning",
        fmi_target: str = "me",
        fmi_version: str = "2.0",
        platform: str = "auto",
        force_compilation: bool = False,
    ) -> CompilationOperations:
        """Compiles the model to an FMU. Returns an ModelExecutableOperation class
        object.

        Args:
            compiler_options: An compilation options class instance of
                CompilerOptions or a dictionary object containing the
                compiler options.
            runtime_options: An runtime options class instance of
                RuntimeOptions or a dictionary object containing the
                runtime options. Default: None.
            compiler_log_level: The logging for the compiler. Possible values are
                "error", "warning", "info", "verbose" and "debug". Default: 'warning'.
            fmi_target: Compiler target. Possible values are 'me' and 'cs'.
                Default: 'me'.
            fmi_version: The FMI version. Valid options are '1.0' and '2.0'.
                Default: '2.0'.
            platform: Platform for FMU binary.The OS running the Impact server must
                match the environment that runs the compiled FMU. This is necessary as
                the binaries packaged with the FMU are based on the platform generating
                the FMU. For example, if the Impact server is running Linux the binary
                in the downloaded FMU is compiled for Linux. The downloaded FMU can
                then not be simulated on Windows. Default: 'auto'. Supported
                options are:-

                - 'auto': platform is selected automatically
                - "linux64": generate a 32 bit FMU
                - "win64": generate a 64 bit FMU

            force_compilation: Force a model compilation.

        Returns:
            An ModelExecutableOperation class object.

        Example::

            compile_ops=model.compile(compiler_options)
            compile_ops.cancel()
            compile_ops.status
            compiler_options=custom_function.get_compiler_options().with_values(c_compiler='gcc')
            runtime_options={'cs_solver':0}
            model.compile(compiler_options, runtime_options).wait()
            model.compile({'c_compiler':'gcc'}).wait()

        """
        _assert_valid_compilation_options(
            compiler_options=compiler_options, runtime_options=runtime_options
        )
        compiler_options = (
            dict(compiler_options)
            if isinstance(compiler_options, CompilerOptions)
            else compiler_options
            if isinstance(compiler_options, dict)
            else {}
        )
        runtime_options = (
            dict(runtime_options)
            if isinstance(runtime_options, RuntimeOptions)
            else runtime_options
            if isinstance(runtime_options, dict)
            else {}
        )

        body = {
            "input": {
                "class_name": self._class_name,
                "compiler_options": compiler_options,
                "runtime_options": runtime_options,
                "compiler_log_level": compiler_log_level,
                "fmi_target": fmi_target,
                "fmi_version": fmi_version,
                "platform": platform,
            }
        }
        if not force_compilation:
            fmu_id, modifiers = self._sal.model_executable.fmu_setup(
                self._workspace_id, body, True
            )
            if fmu_id:
                return CachedModelExecutableOperation[ModelExecutable](
                    self._workspace_id,
                    fmu_id,
                    self._sal,
                    ModelExecutable.from_operation,
                    None,
                    modifiers,
                )

        # No cached FMU, setup up a new one
        fmu_id, _ = self._sal.model_executable.fmu_setup(
            self._workspace_id, body, False
        )

        return ModelExecutableOperation[ModelExecutable](
            self._workspace_id,
            self._sal.model_executable.compile_model(self._workspace_id, fmu_id),
            self._sal,
            ModelExecutable.from_operation,
        )

    def new_experiment_definition(
        self,
        custom_function: CustomFunction,
        *,
        fmi_target: str = "me",
        fmi_version: str = "2.0",
        platform: str = "auto",
        compiler_log_level: str = "warning",
        compiler_options: Optional[CompilerOptionsOrDict] = None,
        runtime_options: Optional[RuntimeOptionsOrDict] = None,
        solver_options: Optional[SolverOptionsOrDict] = None,
        simulation_options: Optional[SimulationOptionsOrDict] = None,
        simulation_log_level: str = "WARNING",
        initialize_from: Optional[CaseOrExperimentOrExternalResult] = None,
    ) -> SimpleModelicaExperimentDefinition:
        """Returns a new experiment definition using this Model.

        Args:
            custom_function: The custom function to use for this experiment.
            fmi_target: Compiler target. Possible values are 'me' and 'cs'.
            Default: 'me'.
            fmi_version: The FMI version. Valid options are '1.0' and '2.0'.
            Default: '2.0'.
            platform: Platform for FMU binary.The OS running the Impact server must
                match the environment that runs the compiled FMU. This is necessary as
                the binaries packaged with the FMU are based on the platform generating
                the FMU. For example, if the Impact server is running Linux the binary
                in the downloaded FMU is compiled for Linux. The downloaded FMU can
                then not be simulated on Windows. Default: 'auto'. Supported
                options are:-

                - 'auto': platform is selected automatically
                - "linux64": generate a 32 bit FMU
                - "win64": generate a 64 bit FMU
            compiler_log_level: The logging for the compiler. Possible values are
                "error", "warning", "info", "verbose" and "debug". Default: 'warning'.
            compiler_options: An compilation options class instance of
                CompilerOptions or a dictionary object containing the compiler options.
            runtime_options: An runtime options class instance of RuntimeOptions or
                a dictionary object containing the runtime options. Default: None.
            solver_options: The solver options to use for this experiment.
                By default the options is set to None, which means the default options
                for the custom_function input is used.
            simulation_options:
                The simulation_options to use for this experiment. By default the
                options is set to None, which means the default options for the
                custom_function input is used.
            simulation_log_level:
                Simulation log level for this experiment. Default is 'WARNING'.
            initialize_from: Optional entity to initialize from. An instance of
                Case or Experiment or ExternalResult. Default: None

        Example::

            model = workspace.get_model("Modelica.Blocks.Examples.PID_Controller")
            dynamic = workspace.get_custom_function('dynamic')
            solver_options = {'atol':1e-8}
            simulation_options = dynamic.get_simulation_options().
            with_values(ncp=500)
            experiment_definition = model.new_experiment_definition(
                dynamic,
                solver_options=solver_options,
                simulation_options=simulation_options
            )
            experiment = workspace.execute(experiment_definition).wait()

        """
        options = self._sal.project.project_options_get(
            self._project_id, self._workspace_id, custom_function=custom_function.name
        )
        project_options = ProjectExecutionOptions(options, custom_function.name)
        return SimpleModelicaExperimentDefinition(
            model=self,
            custom_function=custom_function,
            compiler_options=compiler_options or project_options.compiler_options,
            fmi_target=fmi_target,
            fmi_version=fmi_version,
            platform=platform,
            compiler_log_level=compiler_log_level,
            runtime_options=runtime_options or project_options.runtime_options,
            solver_options=solver_options or project_options.solver_options,
            simulation_options=simulation_options or project_options.simulation_options,
            simulation_log_level=simulation_log_level,
            initialize_from=initialize_from,
        )

    def import_fmu(
        self,
        fmu_path: str,
        class_name: Optional[str] = None,
        overwrite: bool = False,
        include_patterns: Optional[Union[str, List[str]]] = None,
        exclude_patterns: Optional[Union[str, List[str]]] = None,
        top_level_inputs: Optional[Union[str, List[str]]] = None,
        step_size: float = 0.0,
    ) -> FMUImportOperation:
        """Uploads a FMU.

        Args:
            fmu_path:
                The path for the FMU to be imported.
            class_name:
                Qualified name of generated class. By default, 'class_name' is
                set to the name of the library followed by a name based
                on the filename of the imported FMU.
            overwrite:
                Determines if any already existing files should be overwritten.
                Default: False.
            include_patterns, exclude_patterns:
                Specifies what variables from the FMU to include and/or exclude in the
                wrapper model. These two arguments are patterns or lists of patterns as
                the same kind as the argument 'filter' for the function
                'get_model_variables' in PyFMI. If both 'include_patterns' and
                'exclude_patterns' are given, then all variables that matches
                'include_patterns' but does not match 'exclude_patterns' are included.
                Derivatives and variables with a leading underscore in the name are
                always excluded.
                Default value: None (which means to include all the variables).
            top_level_inputs:
                Specify what inputs that should be kept as inputs, i.e. with or without
                the input keyword. The argument is a pattern similar to the arguments
                include_patterns and exclude_patterns. Example: If
                top_level_inputs = 'my_inputs*', then all input variables matching the
                pattern 'my_inputs*' will be generated as inputs, and all other inputs
                not matching the pattern as model variables. If top_level_inputs = '',
                then no input is imported as an input.
                Default value: None (which means all inputs are kept as inputs)
                Type: str or a list of strings
            step_size:
                Specify what value to set for the parameter for step size in the
                generated model. By default the parameter is set to zero, which
                means that the step size will be set during simulation based
                on simulation properties such as the time interval.
                This can also be manually set to any real non-negative number.
                The value of the step size parameter can also be set via the function
                set_step_size, which must be invoked before importing the model.
                Default value: 0.0 (which during simulation is set according to the
                description above).

        Example::

            workspace = client.get_workspace("test")
            package = workspace.get_model('MyPackage')
            fmu_content = package.import_fmu('C:/A.fmu').wait()
            fmu_content = package.import_fmu('C:/B.fmu',
                class_name="MyPackage.Model").wait()

        """
        resp = self._sal.project.project_get(
            self._project_id, vcs_info=False, size_info=False
        )
        project = Project(
            resp["id"],
            resp["definition"],
            resp["projectType"],
            resp["storageLocation"],
            None,
            self._sal,
        )
        modelica_content = project.get_modelica_library_by_name(self._class_name)
        if not modelica_content:
            raise ValueError(
                f"No modelica package with name {self._class_name} exists "
                f"in the Project - '{project.name}'"
            )
        class_name = class_name or ".".join(
            [
                self._class_name,
                os.path.split(fmu_path)[-1].strip(".fmu"),
            ]
        )
        resp = self._sal.project.fmu_import(
            self._project_id,
            modelica_content.id,
            fmu_path,
            class_name,
            overwrite,
            include_patterns,
            exclude_patterns,
            top_level_inputs,
            step_size=step_size,
        )

        return FMUImportOperation[Model](
            resp["data"]["location"],
            self._workspace_id,
            self._project_id,
            self._sal,
            Model.from_operation,
        )

    @classmethod
    def from_operation(cls, operation: BaseOperation[Model], **kwargs: Any) -> Model:
        assert isinstance(operation, FMUImportOperation)
        return cls(**kwargs, service=operation._sal)



import logging
import os
from typing import Optional, Tuple

from modelon.impact.client.credential_manager import CredentialManager
from modelon.impact.client.jupyterhub import exceptions, sal
from modelon.impact.client.sal.context import Context
from modelon.impact.client.sal.uri import URI

logger = logging.getLogger(__name__)


def _get_jupyter_token(credential_manager: CredentialManager, interactive: bool) -> str:
    jupyter_token = credential_manager.get_key(interactive=interactive)
    if not jupyter_token:
        raise exceptions.NoJupyterHubTokenError("No JupyterHub API token is given")

    return jupyter_token


def _running_in_jupyterhub_environment() -> bool:
    return os.environ.get("JUPYTERHUB_SERVICE_PREFIX") is not None


def _running_with_same_jupyterhub_server(uri: URI) -> bool:
    return str(uri) in os.environ.get("IMPACT_URL", "")


def authorize(
    uri: URI,
    interactive: bool,
    context: Optional[Context] = None,
    credential_manager: Optional[CredentialManager] = None,
    service: Optional[sal.JupyterHubService] = None,
) -> Tuple[URI, sal.JupyterContext]:

    jupyter_context = sal.JupyterContext(base=context)
    service = service or sal.JupyterHubService()

    if _running_in_jupyterhub_environment() and _running_with_same_jupyterhub_server(
        uri
    ):
        jupyter_context.token = os.environ.get("JUPYTERHUB_API_TOKEN", "")
        server = os.environ.get("JUPYTERHUB_SERVICE_PREFIX")
        user = service.get_user_data(uri, jupyter_context, server)
        return user.impact_server_uri(uri), jupyter_context

    help_text = f"Enter JupyterHub API token (can be generated at {uri / 'token'}):"
    credential_manager = credential_manager or CredentialManager(
        file_id="jupyterhub-api.key",
        env_names=[
            "JUPYTERHUB_API_TOKEN",
            "MODELON_IMPACT_JUPYTERHUB_CLIENT_API_TOKEN",
        ],
        interactive_help_text=help_text,
    )

    try:
        jupyter_context.token = _get_jupyter_token(credential_manager, interactive)
        user = service.get_user_data(uri, jupyter_context)
    except exceptions.JupyterHubAuthrizationError:
        if interactive:
            logger.warning(
                "Could not authorize with the provided JupyterHub API token, "
                "please enter a new token"
            )
            jupyter_context.token = credential_manager.get_key_from_prompt()
            user = service.get_user_data(uri, jupyter_context)
        else:
            raise

    jupyter_token = jupyter_context.token
    if interactive and jupyter_token:
        credential_manager.write_key_to_file(jupyter_token)

    if not user.server_running():
        uri_to_start_server = uri / "hub/home"
        raise exceptions.NoJupyterHubServerRunningError(
            "The user does not have a server running. "
            f"Go to {uri_to_start_server} and start a server."
        )

    return user.impact_server_uri(uri), jupyter_context



from __future__ import annotations

import enum
import logging
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Optional, Union

from modelon.impact.client.entities.content import ContentType, ProjectContent
from modelon.impact.client.entities.custom_function import CustomFunction
from modelon.impact.client.operations.content_import import ContentImportOperation
from modelon.impact.client.operations.project_import import ProjectImportOperation
from modelon.impact.client.options import ProjectExecutionOptions
from modelon.impact.client.sal.service import Service

if TYPE_CHECKING:
    from modelon.impact.client.operations.base import BaseOperation

logger = logging.getLogger(__name__)
RepoURL = Union["GitRepoURL", "SvnRepoURL"]


@enum.unique
class ProjectType(enum.Enum):
    """Type of project."""

    LOCAL = "LOCAL"
    RELEASED = "RELEASED"
    SYSTEM = "SYSTEM"


@enum.unique
class StorageLocation(enum.Enum):
    """The storage location of the project."""

    APPMODE = "APPMODE"
    USERSPACE = "USERSPACE"
    SYSTEM = "SYSTEM"


@dataclass
class GitRepoURL:
    """GitRepoURL represents a project referenced in a git repository String
    representation is url[@[refname][:sha1]]"""

    url: str
    """URL without protocol part, e.g., gitlab.modelon.com/group/project/repository."""

    refname: str = ""
    """Reference name (branch, tag or similar)"""

    sha1: str = ""
    """Commit hash."""

    def __str__(self) -> str:
        repo_url = self.url
        if self.refname or self.sha1:
            repo_url += "@"
        if self.refname:
            repo_url += self.refname
        if self.sha1:
            repo_url += ":" + self.sha1
        return repo_url

    @classmethod
    def from_dict(cls, data: Any) -> GitRepoURL:
        return cls(
            url=data.get("url"), refname=data.get("refname"), sha1=data.get("sha1")
        )


@dataclass
class SvnRepoURL:
    """SvnRepoURL represents a project referenced in a Subversion repository String
    representation is url/trunk/subdir[@[rev]]"""

    root_url: str
    """URL without protocol part up to branch part, e.g., svn.modelon.com/PNNN/"""

    branch: str = ""
    """Non-empty if it's standard layout and can be either trunk or branches/name or
    tags/name."""

    url_from_root: str = ""
    """URL segment after branch (could be saved in subdir as well)"""

    rev: str = ""
    """Revision number or empty (means HEAD)"""

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, SvnRepoURL):
            return False
        return (
            self.get_rev() == other.get_rev()
            and self.root_url == other.root_url
            and self.branch == other.branch
            and self.url_from_root == other.url_from_root
        )

    def get_rev(self) -> str:
        rev = self.rev
        if rev == "":
            return "HEAD"
        return rev

    def __str__(self) -> str:
        segments = [self.root_url]
        if self.branch:
            segments.append(self.branch)
        if self.url_from_root:
            segments.append(self.url_from_root)
        repo_url = "/".join(segments)
        if self.rev:
            repo_url += "@" + self.rev
        return repo_url

    @classmethod
    def from_dict(cls, data: Any) -> SvnRepoURL:
        return cls(
            root_url=data.get("rootUrl"),
            branch=data.get("branch"),
            url_from_root=data.get("urlFromRoot"),
            rev=data.get("rev"),
        )


@dataclass
class VcsUri:
    service_kind: str
    service_url: str
    repourl: RepoURL
    protocol: str
    subdir: str

    def __str__(self) -> str:
        uri = f"{self.service_kind.lower()}+{self.protocol}://{self.repourl}"
        subdir = self.subdir
        if subdir not in ["", "."]:
            uri = uri + "#" + subdir
        return uri

    @classmethod
    def from_dict(cls, data: Any) -> VcsUri:
        repo_url = data.get("repoUrl")
        service_kind = data.get("serviceKind")
        return cls(
            service_kind=service_kind,
            service_url=data.get("serviceUrl"),
            repourl=GitRepoURL.from_dict(repo_url)
            if service_kind.lower() == "git"
            else SvnRepoURL.from_dict(repo_url),
            protocol=data.get("protocol"),
            subdir=data.get("subdir"),
        )


class ProjectDependency:
    """Dependency entry for a project."""

    def __init__(self, data: Dict[str, Any]):
        self._data = data

    @property
    def name(self) -> Optional[str]:
        """The name of the project dependency."""
        return self._data.get("name")

    @property
    def version_specifier(self) -> Optional[str]:
        """Version specifier."""
        return self._data.get("versionSpecifier")


class ProjectDefinition:
    """Impact project definition."""

    def __init__(self, data: Dict[str, Any]):
        self._data = data

    @property
    def name(self) -> str:
        return self._data["name"]

    @property
    def version(self) -> Optional[str]:
        return self._data.get("version")

    @property
    def format(self) -> str:
        return self._data["format"]

    @property
    def dependencies(self) -> List[ProjectDependency]:
        dependencies = self._data.get("dependencies", [])
        return [ProjectDependency(dependency) for dependency in dependencies]

    @property
    def content(self) -> list:
        return self._data.get("content", [])

    @property
    def execution_options(self) -> List[ProjectExecutionOptions]:
        execution_options = self._data.get("executionOptions", [])
        return [
            ProjectExecutionOptions(
                execution_option, execution_option["customFunction"]
            )
            for execution_option in execution_options
        ]

    def to_dict(self) -> Dict[str, Any]:
        return self._data


class Project:
    """Class containing Project functionalities."""

    def __init__(
        self,
        project_id: str,
        project_definition: Union[ProjectDefinition, Dict[str, Any]],
        project_type: Union[ProjectType, str],
        storage_location: Union[StorageLocation, str],
        vcs_uri: Optional[VcsUri],
        service: Service,
    ):
        self._project_id = project_id
        self._project_definition = (
            ProjectDefinition(project_definition)
            if isinstance(project_definition, dict)
            else project_definition
        )
        self._vcs_uri = vcs_uri or None
        self._project_type = (
            ProjectType(project_type) if isinstance(project_type, str) else project_type
        )
        self._storage_location = (
            StorageLocation(storage_location)
            if isinstance(storage_location, str)
            else storage_location
        )
        self._sal = service

    def __repr__(self) -> str:
        return f"Project with id '{self._project_id}'"

    def __eq__(self, obj: object) -> bool:
        return isinstance(obj, Project) and obj._project_id == self._project_id

    @property
    def id(self) -> str:
        """Project id."""
        return self._project_id

    @property
    def name(self) -> str:
        """Project name."""
        return self.definition.name

    @property
    def size(self) -> float:
        """Project size in bytes."""
        return self._sal.project.project_get(self.id, vcs_info=False, size_info=True)[
            "size"
        ]

    @property
    def definition(self) -> ProjectDefinition:
        return self._project_definition

    @property
    def vcs_uri(self) -> Optional[VcsUri]:
        """Project vcs uri."""
        return self._vcs_uri

    @property
    def project_type(self) -> ProjectType:
        return self._project_type

    @property
    def storage_location(self) -> StorageLocation:
        return self._storage_location

    def delete(self) -> None:
        """Deletes a project.

        Example::

            project.delete()

        """
        self._sal.project.project_delete(self._project_id)

    def _get_project_content(self, content: Dict[str, str]) -> ProjectContent:
        return ProjectContent(content, self._project_id, self._sal)

    def get_contents(self) -> Iterable[ProjectContent]:
        """Get project contents.

        Example::

            project.get_contents()

        """
        return [
            self._get_project_content(content)
            for content in self._project_definition.content
        ]

    def get_content(self, content_id: str) -> ProjectContent:
        """Gets the project content with the given ID.

        Args:
            content_id: The ID of the workspace.

        Example::

            project.get_content("79sd8-3n2a4-e3t24")

        """
        resp = self._sal.project.project_content_get(self.id, content_id)
        return ProjectContent(resp, self.id, self._sal)

    def get_content_by_name(
        self, name: str, content_type: Optional[ContentType] = None
    ) -> Optional[ProjectContent]:
        """Gets the first matching project content with the given name.

        Args:
            name: The name of the content.

            content_type: The type of the project content.

        Example::

            from modelon.impact.client import ContentType
            project.get_content_by_name(name, ContentType.MODELICA)

        """
        contents = self.get_contents()
        if content_type:
            contents = (c for c in contents if c.content_type == content_type)
        try:
            return next(c for c in contents if c.name == name)
        except StopIteration:
            return None

    def get_modelica_library_by_name(self, name: str) -> Optional[ProjectContent]:
        """Gets the first matching Modelica library with the given name.

        Args:

            name: The Modelica library name.

        Example::

            project.get_content_by_name(name)

        """
        return self.get_content_by_name(name, ContentType.MODELICA)

    def import_content(
        self, path_to_content: str, content_type: ContentType
    ) -> ContentImportOperation:
        """Upload content to a project.

        Args:

            path_to_content: The path for the content to be imported.

            content_type: The type of the imported content.

        Example::

            from modelon.impact.client import ContentType

            project.import_content('/home/test.mo', ContentType.MODELICA).wait()

        """
        resp = self._sal.project.project_content_upload(
            path_to_content, self._project_id, content_type.value
        )
        return ContentImportOperation[ProjectContent](
            resp["data"]["location"], self._sal, ProjectContent.from_operation
        )

    def import_modelica_library(self, path_to_lib: str) -> ContentImportOperation:
        """Uploads/adds a non-encrypted Modelica library or a Modelica model to the
        project.

        Args:
            path_to_lib:
                The path for the library to be imported. Only '.mo' or '.zip' file
                extension are supported for upload.

        Example::

            content  = project.import_modelica_library('A.mo').wait()
            content  = project.import_modelica_library('B.zip').wait()

        """
        if Path(path_to_lib).suffix.lower() not in [".mo", ".zip"]:
            raise ValueError(
                "Only '.mo' or '.zip' file extension are supported for uploading "
                "Modelica content into project."
            )
        return self.import_content(path_to_lib, ContentType.MODELICA)

    def get_options(
        self, custom_function: CustomFunction, use_defaults: Optional[bool] = False
    ) -> ProjectExecutionOptions:
        """Get project execution option.

        Args:
            custom_function: The CustomFunction class object.
            use_defaults: If true, default options are used.

        Example::

            dynamic = workspace.get_custom_function('dynamic')
            project.get_options(dynamic)

        """
        if use_defaults:
            options = self._sal.project.project_default_options_get(
                custom_function._workspace_id, custom_function.name
            )
        else:
            options = self._sal.project.project_options_get(
                self._project_id, custom_function._workspace_id, custom_function.name
            )
        return ProjectExecutionOptions(options, custom_function.name)

    @classmethod
    def from_operation(
        cls, operation: BaseOperation[Project], **kwargs: Any
    ) -> Project:
        assert isinstance(operation, ProjectImportOperation)
        return cls(**kwargs, vcs_uri=None, service=operation._sal)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict

from modelon.impact.client import exceptions
from modelon.impact.client.operations.base import (
    AsyncOperation,
    AsyncOperationStatus,
    Entity,
)

if TYPE_CHECKING:
    from modelon.impact.client.operations.base import EntityFromOperation
    from modelon.impact.client.sal.service import Service


class ContentImportOperation(AsyncOperation[Entity]):
    """An operation class for the ProjectContent class."""

    def __init__(
        self, location: str, service: Service, create_entity: EntityFromOperation
    ):
        super().__init__(create_entity)
        self._location = location
        self._sal = service
        self._create_entity = create_entity

    def __repr__(self) -> str:
        return f"Content import operations for id '{self.id}'"

    def __eq__(self, obj: object) -> bool:
        return (
            isinstance(obj, ContentImportOperation) and obj._location == self._location
        )

    @property
    def id(self) -> str:
        """Content import id."""
        return self._location.split("/")[-1]

    @property
    def name(self) -> str:
        """Return the name of operation."""
        return "Project content import"

    def _info(self) -> Dict[str, Any]:
        return self._sal.imports.get_import_status(self._location)["data"]

    def data(self) -> Entity:
        """Returns a new ProjectContent class instance.

        Returns:
            A ProjectContent class instance.

        """
        info = self._info()
        if info["status"] == AsyncOperationStatus.ERROR.value:
            raise exceptions.IllegalContentImport(
                f"Content import failed! Cause: {info['error'].get('message')}"
            )
        project_id = self._location.split("/")[-3]
        content_id = info["data"]["contentId"]
        resp = self._sal.project.project_content_get(project_id, content_id)
        return self._create_entity(self, content=resp, project_id=project_id)

    @property
    def status(self) -> AsyncOperationStatus:
        """Returns the upload status as an enumeration.

        Returns:
            The AsyncOperationStatus enum. The status can have the enum values
            AsyncOperationStatus.READY, AsyncOperationStatus.RUNNING or
            AsyncOperationStatus.ERROR

        Example::

            project.import_content('path/to/model.mo').status

        """
        return AsyncOperationStatus(self._info()["status"])



from __future__ import annotations

import enum
import logging
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from modelon.impact.client import exceptions
from modelon.impact.client.entities.asserts import assert_variable_in_result
from modelon.impact.client.entities.case import Case
from modelon.impact.client.entities.interfaces.experiment import ExperimentInterface
from modelon.impact.client.entities.status import ExperimentStatus
from modelon.impact.client.operations import experiment
from modelon.impact.client.options import (
    CompilerOptions,
    RuntimeOptions,
    SimulationOptions,
    SolverOptions,
)

if TYPE_CHECKING:
    from modelon.impact.client.operations.base import BaseOperation
    from modelon.impact.client.sal.service import Service

logger = logging.getLogger(__name__)

ScalarValue = Union[float, int, str]


@enum.unique
class _TrajectoryResponseFormat(enum.Enum):
    V1 = "application/vnd.impact.trajectories.v1+json"
    V2 = "application/vnd.impact.trajectories.v2+json"


class ExperimentResultPoint:
    """Value class with the single time instance data in a experiment."""

    def __init__(self, trajectories: List[Dict[str, Any]], variables: List[str]):
        self._trajectories = trajectories
        self._variables = variables

    @property
    def variables(self) -> List[str]:
        """List of variables."""
        return self._variables

    @property
    def cases(self) -> List[str]:
        """List of case ids."""
        return [case["caseId"] for case in self._trajectories]

    def as_lists(self) -> List[Optional[List[Optional[ScalarValue]]]]:
        """Return a list of results per case.

        None indicates that the case was not run or failed. None value
        for a variable indicated that the variable is not present in the
        specific case output.

        Example::

            result_data = result.as_lists()

        """
        return [
            [item["trajectory"][0] if item else None for item in case_data["items"]]
            for case_data in self._trajectories
        ]


@enum.unique
class _Workflow(enum.Enum):
    """Workflow type."""

    FMU_BASED = "FMU_BASED"
    CLASS_BASED = "CLASS_BASED"


def _assert_experiment_is_complete(
    status: ExperimentStatus, operation_name: str = "Operation"
) -> None:
    if status == ExperimentStatus.NOTSTARTED:
        raise exceptions.OperationNotCompleteError.for_operation(operation_name, status)
    elif status == ExperimentStatus.CANCELLED:
        raise exceptions.OperationFailureError.for_operation(operation_name)


class ExperimentRunInfo:
    """Class containing experiment run information."""

    def __init__(
        self,
        status: ExperimentStatus,
        errors: List[str],
        failed: int,
        successful: int,
        cancelled: int,
        not_started: int,
    ):
        self._status = status
        self._errors = errors
        self._failed = failed
        self._successful = successful
        self._cancelled = cancelled
        self._not_started = not_started

    @property
    def status(self) -> ExperimentStatus:
        """Status info for an Experiment.

        .. Note:: For tracing status changes of a running experiment execution,
            the ExperimentOperation status property must be used and not
            Experiment entities run info(ExperimentRunInfo) status property. See
            example for usage.

        Example::

            # To get the case entity run info status
            status = experiment.run_info.status

            # To get the status updates of a running experiment execution
            experiment_ops = experiment.execute()
            status = experiment_ops.status

        """
        return self._status

    @property
    def errors(self) -> List[str]:
        """A list of errors.

        Is empty unless 'status' attribute is 'FAILED'

        """
        return self._errors

    @property
    def successful(self) -> int:
        """Number of cases in experiment that are successful."""
        return self._successful

    @property
    def failed(self) -> int:
        """Number of cases in experiment thar have failed."""
        return self._failed

    @property
    def cancelled(self) -> int:
        """Number of cases in experiment that are cancelled."""
        return self._cancelled

    @property
    def not_started(self) -> int:
        """Number of cases in experiment that have not yet started."""
        return self._not_started


class ExperimentMetaData:
    """Class containing experiment metadata."""

    def __init__(self, meta_data: Dict[str, Any]):
        self._meta_data = meta_data

    @property
    def user_data(self) -> Dict[str, Any]:
        """User data dictionary object attached to experiment, if any."""
        return self._meta_data.get("user_data", {})

    @property
    def label(self) -> Optional[str]:
        """Experiment label."""
        return self._meta_data.get("label")


class Experiment(ExperimentInterface):
    """Class containing Experiment functionalities."""

    def __init__(
        self,
        workspace_id: str,
        exp_id: str,
        service: Service,
        info: Optional[Dict[str, Any]] = None,
    ):
        self._exp_id = exp_id
        self._workspace_id = workspace_id
        self._sal = service
        self._info = info
        self._fmu_info: Optional[Dict[str, Any]] = None

    def __eq__(self, obj: object) -> bool:
        return isinstance(obj, Experiment) and obj._exp_id == self._exp_id

    def __repr__(self) -> str:
        return f"Experiment with id '{self._exp_id}'"

    @property
    def id(self) -> str:
        """Experiment id."""
        return self._exp_id

    def _get_info(self, cached: bool = True) -> Dict[str, Any]:
        if not cached or self._info is None:
            self._info = self._sal.workspace.experiment_get(
                self._workspace_id, self._exp_id
            )

        return self._info

    @property
    def run_info(self) -> ExperimentRunInfo:
        """Experiment run information.

        Returns:
            A ExperimentRunInfo class object.

        Example::

            status = experiment.run_info.status

        """
        run_info = self._get_info(cached=False)["run_info"]

        status = ExperimentStatus(run_info["status"])
        errors = run_info.get("errors", [])
        failed = run_info.get("failed", 0)
        successful = run_info.get("successful", 0)
        cancelled = run_info.get("cancelled", 0)
        not_started = run_info.get("not_started", 0)
        return ExperimentRunInfo(
            status, errors, failed, successful, cancelled, not_started
        )

    @property
    def metadata(self) -> ExperimentMetaData:
        """Experiment metadata.

        Returns:
            A ExperimentMetaData class object.

        Example::

            user_data = experiment.metadata.user_data

        """

        info = self._get_info(cached=False)
        meta_data = info.get("meta_data", {})
        return ExperimentMetaData(meta_data)

    @property
    def info(self) -> Dict[str, Any]:
        """Deprecated, use 'run_info' attribute."""
        logger.warning("This attribute is deprectated, use 'run_info' instead")
        return self._get_info(cached=False)

    def execute(
        self, with_cases: Optional[List[Case]] = None, sync_case_changes: bool = True
    ) -> experiment.ExperimentOperation:
        """Executes an experiment. Returns an ExperimentOperation class object.

        Args:
            with_cases: A list of cases objects to execute.
            sync_case_changes: Boolean specifying if to sync the cases given with the
                'with_cases' argument against the server before executing the
                experiment. Default is True.

        Returns:
            An ExperimentOperation class object.

        Example::

            experiment = workspace.create_experiment(experiment_definition)
            experiment_ops = experiment.execute()
            experiment_ops.cancel()
            experiment_ops.status
            experiment_ops.wait()

            generate_cases = experiment.execute(with_cases=[]).wait()
            cases_to_execute =  generate_cases.get_case('case_2')
            experiment = experiment.execute(with_cases=[cases_to_execute]).wait()

        """
        if sync_case_changes and with_cases is not None:
            for case in with_cases:
                case.sync()

        case_ids = [case.id for case in with_cases] if with_cases is not None else None
        return experiment.ExperimentOperation[Experiment](
            self._workspace_id,
            self._sal.experiment.experiment_execute(
                self._workspace_id, self._exp_id, case_ids
            ),
            self._sal,
            Experiment.from_operation,
        )

    def is_successful(self) -> bool:
        """Returns True if the Experiment is done and no cases has failed. Use the
        'run_info' attribute to get more info.

        Returns:
            True, if execution process has completed successfully. False, if
            execution process has failed, is cancelled or still running.

        Example::

            experiment.is_successful()

        """
        return (
            self.run_info.status == ExperimentStatus.DONE
            and self.run_info.failed == 0
            and self.run_info.cancelled == 0
        )

    def get_variables(self) -> List[str]:
        """Returns a list of variables available in the result.

        Returns:
            An list of result variables.

        Raises:
            OperationNotCompleteError if simulation process is in progress.
            OperationFailureError if simulation process has failed or was cancelled.

        Example::

            experiment.get_variables()

        """
        _assert_experiment_is_complete(self.run_info.status, "Simulation")
        return self._sal.experiment.result_variables_get(
            self._workspace_id, self._exp_id
        )

    def get_cases(self) -> List[Case]:
        """Returns a list of case objects for an experiment.

        Returns:
            An list of case objects.

        Example::

            experiment.get_cases()

        """
        resp = self._sal.experiment.cases_get(self._workspace_id, self._exp_id)
        return [
            Case(case["id"], self._workspace_id, self._exp_id, self._sal, case)
            for case in resp["data"]["items"]
        ]

    def get_case(self, case_id: str) -> Case:
        """Returns a case object for a given case_id.

        Args:
            case_id: The case_id for the case.

        Returns:
            An case object.

        Example::

            experiment.get_case('case_1')

        """
        case_data = self._sal.experiment.case_get(
            self._workspace_id, self._exp_id, case_id
        )
        return Case(
            case_data["id"], self._workspace_id, self._exp_id, self._sal, case_data
        )

    def get_cases_with_label(self, case_label: str) -> List[Case]:
        """Returns a list of case objects for an experiment with the label.

        Args:
            case_label: The case_label for the case.

        Returns:
            An list of case objects.

        Example::

            experiment.get_cases_with_label('Cruise condition')

        """
        return [case for case in self.get_cases() if case.meta.label == case_label]

    def _validate_and_fetch_trajectories(
        self,
        variables: List[str],
        only_last_point: bool = False,
        format: _TrajectoryResponseFormat = _TrajectoryResponseFormat.V1,
    ) -> Any:
        if not isinstance(variables, list):
            raise TypeError(
                "Please specify the list of result keys for the trajectories of "
                "intrest!"
            )
        _assert_experiment_is_complete(self.run_info.status, "Simulation")
        assert_variable_in_result(variables, self.get_variables())

        return self._sal.experiment.trajectories_get(
            self._workspace_id, self._exp_id, variables, only_last_point, format.value
        )

    def get_trajectories(self, variables: List[str]) -> Dict[str, Any]:
        """Returns a dictionary containing the result trajectories for a list of result
        variables for all the cases.

        Args:
            variables: A list of result variables to fetch trajectories for.

        Returns:
            A dictionary object containing the result trajectories for all cases.

        Raises:
            OperationNotCompleteError if simulation process is in progress.
            OperationFailureError if simulation process was cancelled.
            TypeError if the variable is not a list object.
            ValueError if trajectory variable is not present in the result.

        Example::

            result = experiment.get_trajectories(['h', 'time'])
            height = result['case_1']['h']
            time = result['case_1']['time']

        """
        response = self._validate_and_fetch_trajectories(
            variables, only_last_point=False, format=_TrajectoryResponseFormat.V1
        )

        if response:
            case_nbrs = range(len(response[0]))
        else:
            case_nbrs = range(len(self.get_cases()))

        return {
            f"case_{j + 1}": {
                variable: response[i][j] for i, variable in enumerate(variables)
            }
            for j in case_nbrs
        }

    def get_last_point(
        self, variables: Optional[List[str]] = None
    ) -> ExperimentResultPoint:
        """Returns a ExperimentResultPoint class for a list of result variables for all
        the cases.

        Args:
            variables: An optional list of result variables to fetch
            trajectories for.

        Returns:
            An ExperimentResultPoint class object.

        Raises:
            OperationNotCompleteError if simulation process is in progress.
            OperationFailureError if simulation process was cancelled.
            TypeError if the variable is not a list object.
            ValueError if trajectory variable is not present in the result.

        Example::

            result = experiment.get_last_point(['h', 'time'])

            # Convert to Pandas data frame
            df = pd.DataFrame(data=result.as_lists(), columns=result.variables,
                index=result.cases)
            df.index.name = "Cases"

        """
        format = _TrajectoryResponseFormat.V2
        if variables is None:
            variables = self.get_variables()
            trajectories = self._sal.experiment.trajectories_get(
                self._workspace_id,
                self._exp_id,
                variables,
                last_point_only=True,
                format=format.value,
            )
        else:
            trajectories = self._validate_and_fetch_trajectories(
                variables, only_last_point=True, format=format
            )
        return ExperimentResultPoint(trajectories["data"]["items"], variables)

    def delete(self) -> None:
        """Deletes an experiment.

        Example::

            experiment.delete()

        """
        self._sal.experiment.experiment_delete(self._workspace_id, self._exp_id)

    def set_label(self, label: str) -> None:
        """Sets a label (string) for an experiment to distinguish it.

        Example::

            experiment.set_label("Engine run with Oil type B")

        """
        self._sal.experiment.experiment_set_label(
            self._workspace_id, self._exp_id, label
        )

    @classmethod
    def from_operation(
        cls, operation: BaseOperation[Experiment], **kwargs: Any
    ) -> Experiment:
        assert isinstance(operation, experiment.ExperimentOperation)
        return cls(**kwargs, service=operation._sal)

    @property
    def custom_function(self) -> str:
        """Returns the custom function name."""
        return self._get_info()["experiment"]["base"]["analysis"]["type"]

    def _get_fmu_info(self, fmu_id: str) -> Dict[str, Any]:
        if self._fmu_info is None:
            self._fmu_info = self._sal.workspace.fmu_get(self._workspace_id, fmu_id)

        return self._fmu_info

    def _get_workflow(self) -> _Workflow:
        model = self._get_info()["experiment"]["base"]["model"]
        return _Workflow.CLASS_BASED if model.get("modelica") else _Workflow.FMU_BASED

    def get_class_name(self) -> str:
        """Return the model class name."""
        model = self._get_info()["experiment"]["base"]["model"]
        if self._get_workflow() == _Workflow.CLASS_BASED:
            return model["modelica"]["className"]
        return self._get_fmu_info(model["fmu"]["id"])["input"]["class_name"]

    def get_compiler_options(self) -> CompilerOptions:
        """Return a CompilerOptions object."""
        model = self._get_info()["experiment"]["base"]["model"]
        if self._get_workflow() == _Workflow.CLASS_BASED:
            return CompilerOptions(
                model["modelica"].get("compilerOptions", {}), self.custom_function
            )
        fmu_info = self._get_fmu_info(model["fmu"]["id"])["input"]
        return CompilerOptions(
            fmu_info.get("compiler_options", {}), self.custom_function
        )

    def get_runtime_options(self) -> RuntimeOptions:
        """Return a RuntimeOptions object."""
        model = self._get_info()["experiment"]["base"]["model"]
        if self._get_workflow() == _Workflow.CLASS_BASED:
            return RuntimeOptions(
                model["modelica"].get("runtimeOptions", {}), self.custom_function
            )
        fmu_info = self._get_fmu_info(model["fmu"]["id"])["input"]
        return RuntimeOptions(fmu_info.get("runtime_options", {}), self.custom_function)

    def get_simulation_options(self) -> SimulationOptions:
        """Return a SimulationOptions object."""
        analysis = self._get_info()["experiment"]["base"]["analysis"]
        return SimulationOptions(
            analysis.get("simulationOptions", {}), self.custom_function
        )

    def get_solver_options(self) -> SolverOptions:
        """Return a SolverOptions object."""
        analysis = self._get_info()["experiment"]["base"]["analysis"]
        return SolverOptions(analysis.get("solverOptions", {}), self.custom_function)
